// Written by Philipp Erler
// Last change 2012-03-20
// This code is not meant to be sold.
// You may use this code or parts of it for your own software.
// Please respect my work. Mentioning my name in your code would be nice.

// Remember to copy the Creative Labs Copyright note if needed.

// This file contains parts from the streaming and enumeration OpenAL samples amongs others
// So I must add the Creative Labs Copyright note:
/*
 * Copyright (c) 2006, Creative Labs Inc.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided
 * that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright notice, this list of conditions and
 * 	     the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright notice, this list of conditions
 * 	     and the following disclaimer in the documentation and/or other materials provided with the distribution.
 *     * Neither the name of Creative Labs Inc. nor the names of its contributors may be used to endorse or
 * 	     promote products derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <windows.h>
#include <stdio.h>
#include <iostream>
#include <conio.h>
#include <math.h>

#include "aldlist.h"

#include "LoadOAL.h"

OPENALFNTABLE ALFunction;

//#define	TEST_WAVE_FILE			"Media\\sine440.wav"
#define	SERVICE_UPDATE_PERIOD	(20)

#define PI						3.1415926536f

#include "OpenALSurroundMachine.h"

using namespace std;

// Tested 8 channel-buffer on a 5.1 Sound Device without real OpenAl support:
// Channel 0 -> Left front
// Channel 1 -> Right front
// Channel 2 -> Left AND Right front
// Channel 3 -> Left AND Right front
// Channel 4 -> Left front
// Channel 5 -> Right front
// Channel 6 -> Front AND Subwoofer
// Channel 7 -> Front AND Subwoofer
	
OpenALSurroundMachine::OpenALSurroundMachine(int frequency, int samples)
{
	// fill properties
	this->frequency = frequency;
	this->samples   = samples;
	this->sourceDist = 1.0f;
	
	// Open AL init
	g_pWaveLoader = new CWaves();
	
	this->PrintDevices();
	this->SelectDevice();
	this->SelectBufferType();
	this->SetListener();
	
	// 3D distance only work with mono sound sources
	this->numElements = samples;
	//this->numElements = this->channels * samples;
	//this->numElements = samples;
	this->dataSize = sizeof(short) * numElements;
	for (int i = 0; i < NUMBUFFERS; ++i)
		this->pData[i] = new short[numElements];

	/*
	// Load Wave file into OpenAL Buffer
	if (!LoadWaveToBuffer((char*)(TEST_WAVE_FILE), uiBuffer))
	{
		printf("Failed to load %s\n", (TEST_WAVE_FILE));
	}
	*/
}

OpenALSurroundMachine::~OpenALSurroundMachine(void)
{
	// Clean up by deleting Source(s) and Buffer(s)
	alSourceStop(uiSource);
    alDeleteSources(1, &uiSource);
	alDeleteBuffers(1, uiBuffers);
	
	ALCcontext *pContext;
	ALCdevice *pDevice;

	pContext = alcGetCurrentContext();
	pDevice = alcGetContextsDevice(pContext);
	
	alcMakeContextCurrent(NULL);
	alcDestroyContext(pContext);
	alcCloseDevice(pDevice);
	
	if (g_pWaveLoader)
	{
		delete g_pWaveLoader;
		g_pWaveLoader = NULL;
	}
}

void OpenALSurroundMachine::SetListener()
{
	ALfloat listenerPos[]={0.0,0.0,0.0};				//at origin
	ALfloat listenerVel[]={0.0,0.0,0.0};				//not moving
	ALfloat listenerOri[]={0.0,0.0,-1.0, 0.0,1.0,0.0};	//looking into the monitor, standing on feet

	ALenum error;

	// Position ...
	alListenerfv(AL_POSITION, listenerPos);
	if ((error = alGetError()) != AL_NO_ERROR)
	{
		//DisplayALError("alListenerfv POSITION : ", error);
		return;
	}
	// Velocity ...
	alListenerfv(AL_VELOCITY, listenerVel);
	if ((error = alGetError()) != AL_NO_ERROR)
	{
		//DisplayALError("alListenerfv VELOCITY : ", error);
		return;
	}
	// Orientation ...
	alListenerfv(AL_ORIENTATION, listenerOri);
	if ((error = alGetError()) != AL_NO_ERROR)
	{
		//DisplayALError("alListenerfv ORIENTATION : ", error);
		return;
	}
}

void OpenALSurroundMachine::SetSource()
{
	ALfloat sourcePos[3];
	
	sourcePos[0] = sourceDist * sin((float)sourceAngle*PI/180.0f);		//x
	sourcePos[1] = 0.0f;						//y = 0
	sourcePos[2] = sourceDist * cos((float)sourceAngle*PI/180.0f);		//z

	alSourcefv(uiSource, AL_POSITION, sourcePos);
}

short* OpenALSurroundMachine::CreateBufferData()
{	
	for (int iBuffer = 0; iBuffer < NUMBUFFERS; ++iBuffer) 
	{
		for (int iSample = 0; iSample < samples; ++iSample) 
		{
			// 3D distance only work with mono sound sources
			pData[iBuffer][iSample] = DataFunction(iSample, 0, frequency);
			/*
			for (int iChannel = 0; iChannel < channels; ++iChannel)
			{
				//pData[iBuffer][iSample*channels + iChannel] = DataFunction(iSample, iChannel, frequency);
			}
			*/
		}
	}
	return pData[0];
}

void OpenALSurroundMachine::PrintDevices( void )
{
    ALint iDeviceIndex = 0;
    const ALCchar* pDeviceNames = NULL;
    const ALCchar* pDefaultDevice = NULL;

	if( LoadOAL10Library( NULL, &ALFunction ) )
    {
        iDeviceIndex = 0;

        printf("\nDevices available:\n");
        printf("============================================\n");
        printf("     %-40s   OpenAL Spec/Ext Support\n", "           ");
        printf("#    %-40s  Default  1.0   1.1   EFX\n\n", "Device Name");
        if( ALFunction.alcIsExtensionPresent( NULL, "ALC_ENUMERATE_ALL_EXT" ) )
        {
            pDeviceNames = ALFunction.alcGetString( NULL, ALC_ALL_DEVICES_SPECIFIER );
            pDefaultDevice = ALFunction.alcGetString( NULL, ALC_DEFAULT_ALL_DEVICES_SPECIFIER );
            if( !strlen(pDeviceNames) )
            {
                printf("None found.\n");
            }
            else
            {
                while( pDeviceNames && *pDeviceNames )
                {
                    printf("[%-2d] - %-40s", iDeviceIndex, pDeviceNames );

                    if( strlen(pDeviceNames) > 40 )
                    {
                        printf("\n%-45s", "" );
                    }

                    // Open and query for extensions
                    ALCdevice* pDevice = ALFunction.alcOpenDevice( pDeviceNames );

                    if( pDevice )
                    {
                        ALint iMajorVersion, iMinorVersion;
                        ALboolean bSpec10Support = AL_FALSE;
                        ALboolean bSpec11Support = AL_FALSE;
                        ALboolean bEFXSupport = AL_FALSE;

                        ALFunction.alcGetIntegerv( pDevice, ALC_MAJOR_VERSION, sizeof(ALint), &iMajorVersion );
                        ALFunction.alcGetIntegerv( pDevice, ALC_MINOR_VERSION, sizeof(ALint), &iMinorVersion );

                        if ( (iMajorVersion == 1) && (iMinorVersion == 0) )
							bSpec10Support = AL_TRUE;
						else if ( (iMajorVersion > 1) || ((iMajorVersion == 1) && (iMinorVersion >= 1)) )
							bSpec11Support = AL_TRUE;

                        if( ALFunction.alcIsExtensionPresent( pDevice, "ALC_EXT_EFX") )
                        {
                            bEFXSupport = AL_TRUE;
                        }

                        // Print default flag
                        printf("   %-8c", !strcmp( pDeviceNames, pDefaultDevice ) ? 'X' : ' ' );
                        // Print Spec version flags
                        printf(" %-5c %-5c", bSpec10Support ? 'X' : ' ', bSpec11Support ? 'X' : ' ' );
                        // Print XRAM ext support flag
                        printf(" %-5c", bEFXSupport ? 'X' : ' ');

                        ALFunction.alcCloseDevice( pDevice );
                    }
                    else
                    {
                        printf("ERROR: Doesn't open!\n");
                    }
                    printf("\n");
                    ++iDeviceIndex;
					devNames.push_back(pDeviceNames);
                    pDeviceNames += strlen(pDeviceNames) + 1;
                }
            }
        }
    }
}

void OpenALSurroundMachine::SelectDevice( void )
{
	//read index of chosen device
	char buf[2+1];
	printf("\nChosen device: ");
    fflush(stdout);
	
    while (fgets(buf, sizeof buf, stdin)) 
	{
		//check for invalid input 
		if (sscanf(buf, "%d", &chosenDev) == 1) 
		{
			//if valid index
			if (chosenDev >= 0 && chosenDev < (int)devNames.size())
			{
				//ALDeviceList *pDeviceList = new ALDeviceList();
				pDevice = ALFunction.alcOpenDevice( devNames[chosenDev] );
				pContext = alcCreateContext(pDevice, NULL);
				
				if (pContext)
				{
					printf("\nOpened Device %s\n", alcGetString(pDevice, ALC_DEVICE_SPECIFIER));
					alcMakeContextCurrent(pContext);

					//Print Extensions
					printf("\nSupported Extensions: ");
					if (alcIsExtensionPresent(pDevice, "ALC_EXT_CAPTURE") == AL_TRUE)
						printf("ALC_EXT_CAPTURE ");
					if (alcIsExtensionPresent(pDevice, "ALC_EXT_EFX") == AL_TRUE)
						printf("ALC_EXT_EFX ");
					if (alcIsExtensionPresent(pDevice, "AL_EXT_OFFSET") == AL_TRUE)
						printf("AL_EXT_OFFSET ");
					if (alcIsExtensionPresent(pDevice, "AL_EXT_LINEAR_DISTANCE") == AL_TRUE)
						printf("AL_EXT_LINEAR_DISTANCE ");
					if (alcIsExtensionPresent(pDevice, "AL_EXT_EXPONENT_DISTANCE") == AL_TRUE)
						printf("AL_EXT_EXPONENT_DISTANCE ");
					if (alcIsExtensionPresent(pDevice, "EAX2.0") == AL_TRUE)
						printf("EAX2.0 ");
					if (alcIsExtensionPresent(pDevice, "EAX3.0") == AL_TRUE)
						printf("EAX3.0 ");
					if (alcIsExtensionPresent(pDevice, "EAX4.0") == AL_TRUE)
						printf("EAX4.0 ");
					if (alcIsExtensionPresent(pDevice, "EAX5.0") == AL_TRUE)
						printf("EAX5.0 ");
					if (alcIsExtensionPresent(pDevice, "EAX-RAM") == AL_TRUE)
						printf("EAX-RAM ");
					printf("\n\n");

				}
				else
				{
					alcCloseDevice(pDevice);
				}
				//delete pDeviceList;
				break;
			}
			printf("\nChosen device: ");
			fflush(stdout);
		}
	}
}

void OpenALSurroundMachine::SelectBufferType(void)
{
	char buf[1+1];

	//read index of chosen device
	printf("\n[0] Mono");
	printf("\n[1] Stereo");
	printf("\n[2] Quadrophonic");
	printf("\n[3] 5.1 Surround");
	printf("\n[4] 6.1 Surround");
	printf("\n[5] 7.1 Surround");
	printf("\nEnter sound buffer format: ");

	int bufferFormat = -1;
    fflush(stdout);
    while (fgets(buf, sizeof buf, stdin)) 
	{
		//invalid input 
		if (sscanf(buf, "%d", &bufferFormat) == 1) 
		{
			//if valid index
			if (bufferFormat >= 0 && bufferFormat <= 5)
			{
				//set number of channels
				switch(bufferFormat)
				{
					case 0: {this->channels = 1; break;}
					case 1: {this->channels = 2; break;}
					case 2: {this->channels = 4; break;}
					case 3: {this->channels = 6; break;}
					case 4: {this->channels = 7; break;}
					case 5: {this->channels = 8; break;}
				}
				break;
			}
			printf("\nEnter sound buffer format: ");
			fflush(stdout);
		}
	}
}

void OpenALSurroundMachine::CreateBuffer(short (*BufferFunction)(int sample, int channel, int frequency))
{
	this->DataFunction = BufferFunction;

	// Generate an AL Buffer
	alGenBuffers( NUMBUFFERS, uiBuffers );

	// Generate a Source to playback the Buffer
    alGenSources( 1, &uiSource );

	for (int iBuffer = 0; iBuffer < NUMBUFFERS; ++iBuffer)
	{
		// OpenAL implementations are only guaranteed to support MONO and STEREO Buffers, so we need
		// to check if this implementation supports multi-channel formats.
		// This is done by making alGetEnumValue calls to request the value of the Buffer Format Tag Enum (that will be
		// passed to an alBufferData call).   Enum Values are retrieved by string names.  The following names are defined
		// for multi-channel wave formats ...
		//  AL_FORMAT_MONO16   (16-bit mono)
		//  AL_FORMAT_STEREO16 (16-bit stereo)
		// "AL_FORMAT_QUAD16"	: 4 Channel,   16 bit data
		// "AL_FORMAT_51CHN16"	: 5.1 Channel, 16 bit data
		// "AL_FORMAT_61CHN16"	: 6.1 Channel, 16 bit data
		// "AL_FORMAT_71CHN16"	: 7.1 Channel, 16 bit data

		// 3D distance only work with mono sound sources
		this->bufferFormat = alGetEnumValue("AL_FORMAT_MONO16");
		/*
		//get enum value
		switch(this->channels)
		{
			case 1: {this->bufferFormat = alGetEnumValue("AL_FORMAT_MONO16");   break;}
			case 2: {this->bufferFormat = alGetEnumValue("AL_FORMAT_STEREO16"); break;}
			case 4: {this->bufferFormat = alGetEnumValue("AL_FORMAT_QUAD16");   break;}
			case 6: {this->bufferFormat = alGetEnumValue("AL_FORMAT_51CHN16");  break;}
			case 7: {this->bufferFormat = alGetEnumValue("AL_FORMAT_61CHN16");  break;}
			case 8: {this->bufferFormat = alGetEnumValue("AL_FORMAT_71CHN16");  break;}
		}
		*/

		alBufferData(uiBuffers[iBuffer], bufferFormat, CreateBufferData(), dataSize, frequency);
		alSourceQueueBuffers(uiSource, 1, &uiBuffers[iBuffer]);
		if (alGetError() != AL_NO_ERROR)
		{
			// When a multi-channel buffer is attached to a Source (either directly using AL_BUFFER) or
			// via Source Queuing ... the call will fail if the playback device does not have enough
			// resources to play the buffer.  (On H/W devices each channel of a Buffer requires a H/W voice,
			// so to play a 5.1 Buffer on a Source requires 6 H/W Voices).
			break;
		}
	}
}
	
void OpenALSurroundMachine::PlaySource(void)
{
	const ALfloat DIST_CHANGE = 0.2f;

	int iTotalBuffersProcessed = 0;

	this->sourceAngle = 0;

	//refresh display
	RefreshDisplay();

	// Start playing source
	alSourcePlay(uiSource);

    fflush(stdout);
	// while no key is pressed
	while (true)
	{
		Sleep( SERVICE_UPDATE_PERIOD );

		if(_kbhit())
		{
			// stop when escape is pressed
			int key = _getch();
			//escape
			if(key == 27)					
			{
				alSourceStop(uiSource);
				break;
			}
			//volume
			//up arrow key
			if (key == 72)
			{
				//increase volume
				if (sourceDist <= 10.0f - DIST_CHANGE)
					sourceDist += DIST_CHANGE;
				SetSource();
				RefreshDisplay();
			}
			//down arrow key
			if (key == 80)
			{
				//decrease volume
				if (sourceDist >= 0.0f + DIST_CHANGE)
					sourceDist -= DIST_CHANGE;
				SetSource();
				RefreshDisplay();
			}
			//right arrow key
			if (key == 77)
			{
				//rotate clockwise for 15 degree
				this->sourceAngle += 15;
				//eliminate inaccuracy and keep in (-2Pi, +2Pi)
				if (this->sourceAngle >= 360) this->sourceAngle = 0;
				SetSource();
				RefreshDisplay();
			}
			//left arrow key
			if (key == 75)
			{
				//rotate counter-clockwise for 15 degree
				this->sourceAngle -= 15;
				//eliminate inaccuracy and keep in (-2Pi, +2Pi)
				if (this->sourceAngle <= -360) this->sourceAngle = 0;
				SetSource();
				RefreshDisplay();
			}
		}

		// Request the number of OpenAL Buffers have been processed (played) on the Source
		int iBuffersProcessed = 0;
		alGetSourcei(uiSource, AL_BUFFERS_PROCESSED, &iBuffersProcessed);
						
		// Keep a running count of number of buffers processed (for logging purposes only)
		iTotalBuffersProcessed += iBuffersProcessed;
		//printf("Buffers Processed %d\r", iTotalBuffersProcessed);

		// For each processed buffer, remove it from the Source Queue, read next chunk of audio
		// data from disk, fill buffer with new data, and add it to the Source Queue
		while (iBuffersProcessed)
		{
			// Remove the Buffer from the Queue.  (uiBuffer contains the Buffer ID for the unqueued Buffer)
			ALuint uiBuffer = 0;
			alSourceUnqueueBuffers(uiSource, 1, &uiBuffer);
			
			alBufferData(uiBuffer, this->bufferFormat, CreateBufferData(), dataSize, frequency);
			alSourceQueueBuffers(uiSource, 1, &uiBuffer);

			iBuffersProcessed--;
		}

		// Check the status of the Source.  If it is not playing, then playback was either completed,
		// or the Source was starved of audio data and needs to be restarted.
		alGetSourcei(uiSource, AL_SOURCE_STATE, &iState);
		if (iState != AL_PLAYING)
		{
			ALint iQueuedBuffers;
			// If there are Buffers in the Source Queue then the Source was starved of audio
			// data, so needs to be restarted (because there is more audio data to play)
			alGetSourcei(uiSource, AL_BUFFERS_QUEUED, &iQueuedBuffers);
			if (iQueuedBuffers)
			{
				alSourcePlay(uiSource);
			}
			else
			{
				// Finished playing
				break;
			}
		}
	}

	printf("\n");
}

void OpenALSurroundMachine::RefreshDisplay(void)
{
	//clear screen
	clrscr();

	//Print Device name
	gotoXY(0, 0);
	printf("%s\n", alcGetString(pDevice, ALC_DEVICE_SPECIFIER));

	//Print Circle
	float drawAngle = 0.0f;
	while(drawAngle <= 2.0f * PI)
	{
		int xPos = (int)(9.0f * cos(drawAngle) + 20.0f);
		int yPos = (int)(6.0f * sin(drawAngle) + 10.0f);

		gotoXY(xPos, yPos);
		putc('.', stdout);
		drawAngle += (PI / 32.0f);
	}

	//Print Listener
	gotoXY(19, 10);
	fputs("[-]", stdout);

	//Print Speakers
	//Front center
	if (this->channels == 1 || this->channels > 5)
	{
		gotoXY(19, 2);
		fputs("(-)", stdout);
	}
	//Front left and right
	if (this->channels > 1)
	{
		gotoXY(13, 3);
		fputs("(-)", stdout);
		gotoXY(25, 3);
		fputs("(-)", stdout);
	}
	//Surround left and right
	if (this->channels > 5)
	{
		gotoXY(7, 10);
		fputs("(-)", stdout);
		gotoXY(31, 10);
		fputs("(-)", stdout);
	}
	//Rear left and right
	if (this->channels == 4 || this->channels == 8)
	{
		gotoXY(13, 17);
		fputs("(-)", stdout);
		gotoXY(25, 17);
		fputs("(-)", stdout);
	}
	//Rear center
	if (this->channels == 7)
	{
		gotoXY(19, 18);
		fputs("(-)", stdout);
	}
	//Subwoofer
	if (this->channels > 5)
	{
		gotoXY(24, 10);
		fputs("(-)", stdout);
	}

	//Print Source
	int xSource = round(9.0f * sin(this->sourceAngle*PI/180.0f) + 19.0f);
	int ySource = round(-6.0f * cos(this->sourceAngle*PI/180.0f) + 10.0f);
	gotoXY(xSource, ySource);
	fputs("(.)", stdout);

	//Print Angle
	gotoXY(0, 19);
	printf("Source Angle: %d degree", sourceAngle);

	// Print Source Position
	gotoXY(0, 20);
	ALfloat sourcePos[3];
	alGetSourcefv(uiSource, AL_POSITION, sourcePos);
	printf("Source Position: x = %f, y = %f, z = %f", sourcePos[0], sourcePos[1], sourcePos[2]);

	//Waiting point
	gotoXY(0, 21);
	fputs("Escape: Close   Left/Right: Move Source   Up/Down: Change Distance\n", stdout);
}

//thanks to: http://www.gamedev.net/topic/308393-how-to-move-cursor-in-windows-console-mode-fixed/
void OpenALSurroundMachine::gotoXY(int x, int y)
{
	COORD coord;
	coord.X = x;
	coord.Y = y;
	SetConsoleCursorPosition(GetStdHandle(STD_OUTPUT_HANDLE), coord);
}

//thanks to: http://www.gamedev.net/topic/308393-how-to-move-cursor-in-windows-console-mode-fixed/
void OpenALSurroundMachine::clrscr()
{
    HANDLE hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
    COORD coord = {0, 0};
    DWORD count;

    CONSOLE_SCREEN_BUFFER_INFO csbi;
    GetConsoleScreenBufferInfo(hStdOut, &csbi);

    FillConsoleOutputCharacter(hStdOut, ' ', csbi.dwSize.X * csbi.dwSize.Y, coord, &count);

    SetConsoleCursorPosition(hStdOut, coord);
}

//thanks to: http://www.codeproject.com/Articles/58289/C-Round-Function
int OpenALSurroundMachine::round(float number)
{
    return (number >= 0.0f) ? (int)(number + 0.5f) : (int)(number - 0.5f);
}


// XRAM functions and enum values

LPEAXSETBUFFERMODE eaxSetBufferMode = NULL;
LPEAXGETBUFFERMODE eaxGetBufferMode = NULL;

ALboolean OpenALSurroundMachine::LoadWaveToBuffer(const char *szWaveFile, ALuint uiBufferID, ALenum eXRAMBufferMode)
{
	WAVEID			WaveID;
	ALint			iDataSize, iFrequency;
	ALenum			eBufferFormat;
	ALchar			*pData;
	ALboolean		bReturn;

	bReturn = AL_FALSE;
	if (g_pWaveLoader)
	{
		if (SUCCEEDED(g_pWaveLoader->LoadWaveFile(szWaveFile, &WaveID)))
		{
			if ((SUCCEEDED(g_pWaveLoader->GetWaveSize(WaveID, (unsigned long*)&iDataSize))) &&
				(SUCCEEDED(g_pWaveLoader->GetWaveData(WaveID, (void**)&pData))) &&
				(SUCCEEDED(g_pWaveLoader->GetWaveFrequency(WaveID, (unsigned long*)&iFrequency))) &&
				(SUCCEEDED(g_pWaveLoader->GetWaveALBufferFormat(WaveID, &alGetEnumValue, (unsigned long*)&eBufferFormat))))
			{
				// Set XRAM Mode (if application)
				if (eaxSetBufferMode && eXRAMBufferMode)
					eaxSetBufferMode(1, &uiBufferID, eXRAMBufferMode);

				alGetError();
				alBufferData(uiBufferID, eBufferFormat, pData, iDataSize, iFrequency);
				if (alGetError() == AL_NO_ERROR)
					bReturn = AL_TRUE;

				g_pWaveLoader->DeleteWaveFile(WaveID);
			}
		}
	}

	return bReturn;
}