// Written by Philipp Erler
// Last change 2012-03-20
// This code is not meant to be sold.
// You may use this code or parts of it for your own software.
// Please respect my work. Mentioning my name in your code would be nice.

#ifndef OPENALSURROUNDMACHINE_H
#define OPENALSURROUNDMACHINE_H

#include <stdio.h>
#include <iostream>
#include <limits.h>
#include <conio.h>

//OpenAl Framework
#include "LoadOAL.h"
#include "CWaves.h"
#include "aldlist.h"

// XRAM Extension function pointer variables and enum values
typedef ALboolean (__cdecl *LPEAXSETBUFFERMODE)(ALsizei n, ALuint *buffers, ALint value);
typedef ALenum    (__cdecl *LPEAXGETBUFFERMODE)(ALuint buffer, ALint *value);

#define NUMBUFFERS				(4)

class OpenALSurroundMachine
{
private:
	int								frequency;
	int								channels;
	int								samples;
	int								numElements;
	int								dataSize;
	short*							pData[NUMBUFFERS];
	std::vector<const ALCchar*>		devNames;
	int								chosenDev;

	// user function to fill buffers
	short							(*DataFunction)(int sample, int channel, int frequency);
	
	ALfloat							sourceDist;
	ALCdevice						*pDevice;
	ALCcontext						*pContext;
	ALuint							uiBuffers[NUMBUFFERS];
	ALuint							uiSource;
	ALuint							uiListener;
	ALint							iState;

	CWaves							*g_pWaveLoader;
	
	ALenum							bufferFormat;
	int								sourceAngle;		//in degree

	void SetListener(void);
	void SetSource(void);

	void RefreshDisplay(void);	

	void gotoXY(int x, int y);
	void clrscr(void);
	int round(float number);

public:
	OpenALSurroundMachine(int frequency, int samples);
	~OpenALSurroundMachine(void);
	
	void CreateBuffer(short (*BufferFunction)(int sample, int channel, int frequency));
	short* CreateBufferData(void);
	void PrintDevices(void);
	void SelectDevice(void);
	void SelectBufferType(void);
	void PlaySource(void);

	ALboolean LoadWaveToBuffer(const char *szWaveFile, ALuint uiBufferID, ALenum eXRAMBufferMode = 0);
};
#endif OPENALSURROUNDMACHINE_H